/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.tokenFifo;

import java.util.Iterator;
import java.util.LinkedList;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

import junit.framework.TestCase;
import junit.textui.TestRunner;

/** 
 * The FIFO run by JUnit framework.
 * 
 * @author ib27688
 *
 */
public class TokenFifoTest extends TestCase {

	public TokenFifoTest(String arg) {
		super(arg);
	}
	
	public void setUp() {
		
	}
	
	public void tearDown() {
		Director.getInstance().masterClear();
	}
	
	public static void main(String[] args) {
		TestRunner.run(TokenFifoTest.class);
	}
	
	// vanilla test for Fifo of capacity 5
	public void testTokenFifo_5_10_10() {
		testTokenFifo(20, 1, 10, 10);
	}
		
	// vanilla test for fifo of capacity 1
	public void testTokenFifo_1_10_10() {
		testTokenFifo(20, 1, 10, 10);
	}
	
	// vanilla test for fifo of capacity 0
	public void testTokenFifo_0_10_10() {
		testTokenFifo(20, 0, 10, 10);
	}
	
	// aggressive source
	public void testTokenFifo_5_0_10() {
		testTokenFifo(40, 5, 0, 10);
	}		

	// aggressive source, capacity 1
	public void testTokenFifo_1_0_10() {
		testTokenFifo(20, 1, 0, 10);
	}		

	// aggressive source, capacity 0
	public void testTokenFifo_0_0_10() {
		testTokenFifo(20, 0, 0, 10);
	}		
	
	// aggressive sink, capacity 5
	public void testTokenFifo_5_10_0() {
		testTokenFifo(40, 5, 10, 0);
	}		

	// aggressive sink, capacity 1
	public void testTokenFifo_1_10_0() {
		testTokenFifo(20, 1, 10, 0);
	}		

	// aggressive sink, capacity 0
	public void testTokenFifo_0_10_0() {
		testTokenFifo(20, 0, 10, 0);
	}		
	
	/**
	 * Test driver
	 * @param n number if items to send
	 * @param capacity fifo capacity
	 * @param srcDelay delay for the source
	 * @param sinkDelay delay for the sink
	 */
	public void testTokenFifo(int n, int capacity, int srcDelay, int sinkDelay) {
		TokenFifoTestGroup test= new TokenFifoTestGroup(n,capacity, srcDelay, sinkDelay);
		Director director= Director.getInstance();
		// run selfTest
		Boolean selfCheck= test.selfCheck();
		assertTrue(selfCheck);
		
		// run the rest of the test only if self-check passed
		if (selfCheck) {
			director.reset();
			// run for long enough to go through all the tests
			director.run();
			boolean testsPassed= test.testCheck();
			assertTrue(testsPassed);
		}
		
	}
	
} // FifoTest

/**
 * Test setup for a token fifo test
 * 
 * @param n - the number of data items the source will produce
 * @param capacity - the capacity of the fifo
 * 
 * @author ib27688
 */
class TokenFifoTestGroup extends CompositeEntity {
	
	int count;
	
	TokenFifoSource src;
	TokenFifoSink sink;
	TokenFifoShip fifo;
	
	public TokenFifoTestGroup(int n, int capacity, int srcDelay, int sinkDelay) {
		super("TokenFifoTestGroup" + n);
		
		count= n;
		
		src= new TokenFifoSource("TokenFifoTestSource", n, srcDelay);
		addMember(src);
		
		fifo= new TokenFifoShip("TestTokenFifo", capacity);
		addMember(fifo);
		
		sink= new TokenFifoSink("TokenFifoTestSink", n, sinkDelay);
		addMember(sink);
		
		// states that connect thet fifo with the source and sink
		State srcDataState= new State("TokenFifoTestDataInState_" + capacity);
		addMember(srcDataState);
		State srcAckDataState= new State("TokenFifoTestAckDataInState_" + capacity);
		addMember(srcAckDataState);
		
		srcDataState.attach(src.getTerminal(FleetGlobal.OUT));
		srcDataState.attach(fifo.getTerminal(FleetGlobal.IN));
		
		srcAckDataState.attach(src.getTerminal(FleetGlobal.ACK_OUT));
		srcAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_IN));
		
		State sinkDataState= new State("TokenFifoTestDataOutState_" + capacity);
		addMember(sinkDataState);
		State sinkAckDataState= new State("TokenFifoTestDataOutState_" + capacity);
		addMember(sinkAckDataState);
		
		sinkDataState.attach(fifo.getTerminal(FleetGlobal.OUT));
		sinkDataState.attach(sink.getTerminal(FleetGlobal.IN));
		
		sinkAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_OUT));
		sinkAckDataState.attach(sink.getTerminal(FleetGlobal.ACK_IN));
		
		// Tokens on the source side
		State srcTokenState= new State("TokenFifoTestSrcTokenState_" + capacity);
		addMember(srcTokenState);
		State srcAckTokenState= new State("TokenFifoTestSrcAckTokenState_" + capacity);
		addMember(srcAckTokenState);
		
		srcTokenState.attach(fifo.getTerminal(FleetGlobal.TOKEN_OUT));
		srcTokenState.attach(src.getTerminal(FleetGlobal.TOKEN_IN));
		
		srcAckTokenState.attach(fifo.getTerminal(FleetGlobal.ACK_TOKEN_OUT));
		srcAckTokenState.attach(src.getTerminal(FleetGlobal.ACK_TOKEN_IN));
		
		// Tokens on the destination side
		State sinkTokenState= new State("TokenFifoTestSinkTokenState_" + capacity);
		addMember(sinkTokenState);
		State sinkAckTokenState= new State("TokenFifoTestSinkAckTokenState_" + capacity);
		addMember(sinkAckTokenState);
		
		sinkTokenState.attach(fifo.getTerminal(FleetGlobal.TOKEN_IN));
		sinkTokenState.attach(sink.getTerminal(FleetGlobal.TOKEN_OUT));
		
		sinkAckTokenState.attach(fifo.getTerminal(FleetGlobal.ACK_TOKEN_IN));
		sinkAckTokenState.attach(sink.getTerminal(FleetGlobal.ACK_TOKEN_OUT));
		
		// take care of initially active states
		srcAckDataState.setInitiallyActive(new Data());
		sinkAckDataState.setInitiallyActive(new Data());
		
		srcTokenState.setInitiallyActive(new Data());
		sinkTokenState.setInitiallyActive(new Data());
		
	} //FifoTestGroup constructor
	
	boolean testCheck() {
		LinkedList<Data> computedResults= new LinkedList<Data>();
		for (int i= 0; i<count; i++) {
			computedResults.addLast(new Data(i));
		}

		// get simulated results
		LinkedList<Data> simulatedResults= sink.getResults();

		/*
		// display computed results 
		System.out.println("\nToken fifo test computed results: ");
		Iterator<Data> ci= computedResults.iterator();
		while (ci.hasNext()) {
			Data d= ci.next();
			System.out.println(d);
		}
		
		// display simulated results		
		System.out.println("\n\nToken fifo test simulated results: ");
		Iterator<Data> si= simulatedResults.iterator();
		while (si.hasNext()) {
			Data d= si.next();
			System.out.println(d);
		}
		*/
		
		return computedResults.equals(simulatedResults);
	} // testCheck
	
} // TokenFifoTestGroup


/**
 * Data source for token fifo under test
 * @author ib27688
 *
 */
class TokenFifoSource extends Component {
	
	Terminal out;
	Terminal ackOut;
	
	Terminal tokenIn;
	Terminal ackTokenIn;
	
	TokenFifoSourceAction srcAction;
	
	public TokenFifoSource(String name, int n, int srcDelay) {
		super(name, null);
		
		srcAction= new TokenFifoSourceAction("FifoSourceAction", n, srcDelay);
		addMember(srcAction);
		out= new OutputTerminal(FleetGlobal.OUT, this);
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
		//attach the terminals
		srcAction.attach(out);
		srcAction.attach(ackOut);
		
		// token terminals
		tokenIn= new InputTerminal(FleetGlobal.TOKEN_IN, this);
		ackTokenIn= new OutputTerminal(FleetGlobal.ACK_TOKEN_IN, this);
		// attach token terminals
		srcAction.attach(tokenIn);
		srcAction.attach(ackTokenIn);
		
	} // FifoSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		else if (name.equals(FleetGlobal.TOKEN_IN)) {
			return tokenIn;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_IN)) {
			return ackTokenIn;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
} // TokenFifoSource


class TokenFifoSourceAction extends TriggerAction {

	int count;
	public TokenFifoSourceAction(String name, int n, int delay) {
		super(name);
		// how many data items to produce
		count= n;
		outputFunctions.put(FleetGlobal.OUT, new GenerateData(delay));
		outputFunctions.put(FleetGlobal.ACK_TOKEN_IN, new AckToken());
		
		//start enabled right away
		enabled= true;
		// enableTrigggerCommand.execute(null);
		
	} // TokenFifoSourceAction
	
	/** Generate new data items */
	private class GenerateData extends Function {

		private int next= 0;
		
		public GenerateData(int delay) {
			super("GenerateData");
			functionDelay= new Delay(delay); // was 20
		}
		
		public Object value() {
			int current= next;
			next++;
			if (next >= count) {
				// done counting, disable the source
				enabled= false;
			}
			if (TokenFifoShipDefaults.sourceDisplayGenerated) {
				director.journal("Source generated " + next);
			}
			
			Payload p= new Payload();
			Data d= new Data(current);
			p.data= d;			
			return p;	
		} // value
	} // GenerateData

	
	/** Generate token acknowledgment */
	private class AckToken extends Function {
		
		public AckToken() {
			super("AckToken");
			functionDelay= TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY_DEF; 
		}
		
		public Object value() {
			return new Data();	
		} // value
	} // PassData

	
} // TokenFifoSourceAction

/**
 * Sink for data for a TokenFifo under test
 * @author ib27688
 *
 */
class TokenFifoSink extends Component {
	// data terminals
	Terminal in;
	Terminal ackIn;
	// token terminals
	Terminal tokenOut;
	Terminal ackTokenOut;
	
	TokenFifoSinkAction sinkAction;
	
	public TokenFifoSink(String name, int n, int sinkDelay) {
		super(name, null);
		
		sinkAction= new TokenFifoSinkAction("FifoSinkAction", n, sinkDelay);
		addMember(sinkAction);

		// data terminals
		in= new InputTerminal(FleetGlobal.IN, this);
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN, this);
		//attach the terminals
		sinkAction.attach(in);
		sinkAction.attach(ackIn);
		
		// token terminals
		tokenOut= new OutputTerminal(FleetGlobal.TOKEN_OUT, this);
		ackTokenOut= new InputTerminal(FleetGlobal.ACK_TOKEN_OUT, this);
		//attach the terminals
		sinkAction.attach(tokenOut);
		sinkAction.attach(ackTokenOut);
		
	} // FifoSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		if (name.equals(FleetGlobal.TOKEN_OUT)) {
			return tokenOut;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_OUT)) {
			return ackTokenOut;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
	/**
	 * Get results...
	 * @return a list of results of tests
	 */
	LinkedList<Data> getResults() {
		return sinkAction.results;
	} // getResults
		
	
} // class TokenFifoSink


class TokenFifoSinkAction extends Action {
	
	// the number or items to receive
	int count;
	
	LinkedList<Data> results;
	
	
	public TokenFifoSinkAction(String name, int n, int delay) {
		super(name);
		// how many data items to receive
		count= n;
		// list of received results
		results= new LinkedList<Data>();
		outputFunctions.put(FleetGlobal.ACK_IN, new ActionFired(delay));
		outputFunctions.put(FleetGlobal.TOKEN_OUT, new GenerateToken());
	} // FifoSinkAction
	
	/** Vacuous function used to acknowldge that the action has fired */ 
	private class ActionFired extends OutputFunction {
		
		public ActionFired(int d) {
			super("Vacous");
			functionDelay= new Delay(d); 
		}
	
		public Data value() {
			Iterator ti= terminals.iterator();
			Terminal t= null;
			boolean done= false;
			while (!done && ti.hasNext()) {
				t= (Terminal)(ti.next());
				if (t.getBaseName() == FleetGlobal.IN) {
					// found the terminal
					done= true;
					Data datum= (Data)t.getData();
					results.addLast(datum);
				}
			}
			if (t != null && TokenFifoShipDefaults.sinkDisplayReceived ) {
				director.journal("Sink received " + ((Data)t.getData()).value);
			}
			return null;
		} // value
	} // ActionFired
	
	/** Generate token  */
	private class GenerateToken extends Function {
		
		public GenerateToken() {
			super("GenerateToken");
			functionDelay= TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY_DEF; 
		}
		
		public Object value() {
			return new Data();	
		} // value
	} // GenerateToken

	
	
} // class TokenFifoSinkAction
